#ifndef _LLOG_H_
#define _LLOG_H_

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>
#include <stdarg.h>
#include <stdbool.h>
#include <time.h>
typedef uint16_t log_mask_t;

#define LOG_MASK_ENDLINE	(1 << 0)
#define LOG_MASK_COLORS		(1 << 1)
#define LOG_MASK_LEVEL		(1 << 2)
#define LOG_MASK_FILE		(1 << 3)
#define LOG_MASK_FUNC		(1 << 4)
#define LOG_MASK_LINE		(1 << 5)
#define LOG_MASK_MODE		(1 << 6)

#define LOG_MASK_SECOND		(1 << 8)
#define LOG_MASK_MINUTE		(1 << 9)
#define LOG_MASK_HOUR		(1 << 10)
#define LOG_MASK_DAY		(1 << 11)
#define LOG_MASK_MONTH		(1 << 12)
#define LOG_MASK_YEAR		(1 << 13)

#define LOG_FILE_SIMPLE		(1 << 15)

#define LOG_MASK_TIME_HMS		(LOG_MASK_SECOND | LOG_MASK_MINUTE | LOG_MASK_HOUR)
#define LOG_MASK_TIME_YMD		(LOG_MASK_DAY | LOG_MASK_MONTH | LOG_MASK_YEAR)
#define LOG_MASK_TIME_ALL		(LOG_MASK_TIME_HMS | LOG_MASK_TIME_YMD)

#define LOG_MASK_ALL_BASE		(LOG_MASK_LEVEL | LOG_MASK_LINE | LOG_MASK_FUNC)
#define LOG_MASK_ALL_MODE		(LOG_MASK_MODE | LOG_MASK_ALL_BASE)
#define LOG_MASK_ALL_FILE		(LOG_MASK_FILE | LOG_MASK_ALL_BASE)

#define LOG_MASK_ALL_DETAIL		(LOG_MASK_MODE | LOG_MASK_FILE | LOG_MASK_ALL_BASE)
#define LOG_MASK_BASE_STDOUT	(LOG_MASK_COLORS | LOG_MASK_ALL_DETAIL | LOG_MASK_TIME_HMS)
#define LOG_MASK_BASE_FILE		(LOG_MASK_ALL_DETAIL | LOG_MASK_TIME_ALL)
#define LOG_MASK_BASE_DEFAULT	(LOG_MASK_BASE_STDOUT | LOG_FILE_SIMPLE)

#ifdef __cplusplus
extern "C" {
#endif

enum { _LOG_TRACE, _LOG_DEBUG, _LOG_INFO, _LOG_WARN, _LOG_ERROR, _LOG_FATAL };

static const char* _log_level_strings[] = {
  "TRACE", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"
};

#define LOG_COLOR(code) \033[##code##m
#define LOG_COLOR_TRACE LOG_COLOR(94)
#define LOG_COLOR_DEBUG LOG_COLOR(36)
#define LOG_COLOR_INFO LOG_COLOR(32)
#define LOG_COLOR_WARN LOG_COLOR(33)
#define LOG_COLOR_ERROR LOG_COLOR(31)
#define LOG_COLOR_FATAL LOG_COLOR(35)

#define _TO_STR(str) #str
#define TO_STR(str) _TO_STR(str)
static const char* _log_level_colors[] = {
	TO_STR(LOG_COLOR_TRACE), TO_STR(LOG_COLOR_DEBUG), TO_STR(LOG_COLOR_INFO),
	TO_STR(LOG_COLOR_WARN), TO_STR(LOG_COLOR_ERROR), TO_STR(LOG_COLOR_FATAL)
};

#define log_head_trace( log, szlog) log_fill_head_str(_LOG_TRACE, log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)
#define log_head_debug( log, szlog) log_fill_head_str(_LOG_DEBUG, log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)
#define log_head_info(  log, szlog) log_fill_head_str(_LOG_INFO,  log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)
#define log_head_warn(  log, szlog) log_fill_head_str(_LOG_WARN,  log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)
#define log_head_error( log, szlog) log_fill_head_str(_LOG_ERROR, log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)
#define log_head_fatal( log, szlog) log_fill_head_str(_LOG_FATAL, log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)
#define log_head(level, log, szlog) log_fill_head_str(level     , log, szlog, LOG_MASK_BASE_DEFAULT, __FILE__, __func__, __func__, __LINE__)

#define log_head_trace_ex( log, szlog, mode, mask) log_fill_head_str(_LOG_TRACE, log, szlog, mask, __FILE__, mode, __func__, __LINE__)
#define log_head_debug_ex( log, szlog, mode, mask) log_fill_head_str(_LOG_DEBUG, log, szlog, mask, __FILE__, mode, __func__, __LINE__)
#define log_head_info_ex(  log, szlog, mode, mask) log_fill_head_str(_LOG_INFO,  log, szlog, mask, __FILE__, mode, __func__, __LINE__)
#define log_head_warn_ex(  log, szlog, mode, mask) log_fill_head_str(_LOG_WARN,  log, szlog, mask, __FILE__, mode, __func__, __LINE__)
#define log_head_error_ex( log, szlog, mode, mask) log_fill_head_str(_LOG_ERROR, log, szlog, mask, __FILE__, mode, __func__, __LINE__)
#define log_head_fatal_ex( log, szlog, mode, mask) log_fill_head_str(_LOG_FATAL, log, szlog, mask, __FILE__, mode, __func__, __LINE__)
#define log_head_ex(level, log, szlog, mode, mask) log_fill_head_str(level     , log, szlog, mask, __FILE__, mode, __func__, __LINE__)

#define _LOG_MAX_HEAD_BUF_SIZE 256

struct _fill_head_parm {
	int level;
	char* buf;
	size_t szbuf;
	log_mask_t mask;
	const char* file;
	const char* mode;
	const char* func;
	int line;

	struct tm* time;
};

static size_t _log_fill_head_time_ymd(struct _fill_head_parm* parm) {
	log_mask_t mask = parm->mask;
	char* buf = parm->buf;
	size_t szbuf = parm->szbuf;
	struct tm* time = parm->time;
	if (LOG_MASK_TIME_YMD & mask) {
		if (LOG_MASK_YEAR & mask) {
			szbuf += strftime(buf + szbuf, _LOG_MAX_HEAD_BUF_SIZE - szbuf, "%Y", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "Y");
		}
		if (LOG_MASK_MONTH & mask) {
			szbuf += strftime(buf + szbuf, _LOG_MAX_HEAD_BUF_SIZE - szbuf, "-%m", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "-M");
		}
		if (LOG_MASK_DAY & mask) {
			szbuf += strftime(buf + szbuf, _LOG_MAX_HEAD_BUF_SIZE - szbuf, "-%d", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "-D");
		}
	}
	return szbuf;
}

static size_t _log_fill_head_time_hms(struct _fill_head_parm* parm) {
	log_mask_t mask = parm->mask;
	char* buf = parm->buf;
	size_t szbuf = parm->szbuf;
	struct tm* time = parm->time;
	if (LOG_MASK_TIME_HMS & mask) {
		if (LOG_MASK_HOUR & mask) {
			szbuf += strftime(buf + szbuf, _LOG_MAX_HEAD_BUF_SIZE - szbuf, "%H", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "h");
		}
		if (LOG_MASK_MINUTE & mask) {
			szbuf += strftime(buf + szbuf, _LOG_MAX_HEAD_BUF_SIZE - szbuf, ":%M", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, ":m");
		}
		if (LOG_MASK_SECOND & mask) {
			szbuf += strftime(buf + szbuf, _LOG_MAX_HEAD_BUF_SIZE - szbuf, ":%S", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, ":s");
		}
	}
	return szbuf;
}



static size_t _log_fill_head_detail(struct _fill_head_parm* parm) {
	log_mask_t mask = parm->mask;
	char* buf = parm->buf;
	size_t szbuf = parm->szbuf;
	int file_offset = 0;

	if (LOG_MASK_COLORS & mask) {
		szbuf += sprintf(buf + szbuf, "%s%-5s\033[0m \033[90m",
			_log_level_colors[parm->level],
			(LOG_MASK_LEVEL & mask) ? _log_level_strings[parm->level] : "nan");
	}
	else {
		szbuf += sprintf(buf + szbuf, "%-5s ",
			(LOG_MASK_LEVEL & mask) ? _log_level_strings[parm->level] : "nan");
	}
	if (LOG_FILE_SIMPLE & mask) {
		for (file_offset = strlen(parm->file);
			file_offset > -1; file_offset -= 1) {
			if (parm->file[file_offset] == '/' || parm->file[file_offset] == '\\') {
				file_offset += 1;
				break;
			}
		}
	}
	szbuf += sprintf(buf + szbuf, "%.128s:%.32s:%.32s:%d:",
		(LOG_MASK_FILE & mask) ? parm->file + file_offset : "file",
		(LOG_MASK_MODE & mask) ? parm->mode : "mode",
		(LOG_MASK_FUNC & mask) ? parm->func : "func",
		(LOG_MASK_LINE & mask) ? parm->line : 0);

	if (LOG_MASK_COLORS & mask) {
		szbuf += sprintf(buf + szbuf, "\033[0m");
	}
	return szbuf;
}

static inline size_t log_fill_head_str(int level, char* log, size_t szlog,
	log_mask_t mask, const char* file, const char* mode, const char* func, int line) {
	size_t szbuf = 0;
	char buf[_LOG_MAX_HEAD_BUF_SIZE] = { 0 };
	
	time_t t = time(NULL);
	struct tm* time = localtime(&t);
	struct _fill_head_parm parm = {
		.level = level,
		.buf = buf,
		.szbuf = 0,
		.mask = mask,
		.file = file,
		.mode = mode,
		.func = func,
		.line = line,

		.time = time
	};
	size_t(*callback[3])(struct _fill_head_parm* parm) = {
		_log_fill_head_time_ymd,
		_log_fill_head_time_hms,
		_log_fill_head_detail
	};
	for (int i = 0; i < sizeof(callback) / sizeof(callback[0]); i++) {
		szbuf += (*callback[i])(&parm);
		parm.buf[szbuf] = '|';
		parm.szbuf = szbuf + 1;
	}
	if (LOG_MASK_ENDLINE & mask) szbuf += sprintf(buf, "\n");
	buf[szbuf] = '\0';
	szbuf += 1;
	if (szlog <= szbuf) return 0;
	strncpy(log, buf, szbuf);
	return szbuf;
}

extern void log_buf(void);

#ifdef __cplusplus
}
#endif

#endif /* _LLOG_H_ */